/*******************************************************************************
 * RGB_Led_Controller.c - Main Code File
 * Autore: Matteo Garia
 * Data Creazione: 19/04/2012
 * Versione: 2.0 (11/11/2012)
 * ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "RGB_Led_Controller.h"

//##############################################################################
// Interrupt Service Routine
//##############################################################################
void interrupt isr(){
//Contatore per la generazione del PWM
    static unsigned char pwmCounter;
//******************************************************************************
// Gestione interrupt TIMER0
//******************************************************************************
    if (TMR0IF)
    {
//Precarico il timer ed azzero il flag di interrupt
        TMR0=TMR0_PRELOAD;
        TMR0IF = 0;
//Conteggio gli interrupt per dare il tick di sistema
        timer0Divider++;
        if(timer0Divider == 200){
            timer0Divider = 0;
            systemTick = 1;
        }
//Generazione PWM
        if (pwm.Left.Red > pwmCounter || pwm.Left.Red == 255){
            PWM_OUT_0 = HIGH;
        }
        else{PWM_OUT_0 = LOW;
        }

        if (pwm.Left.Green > pwmCounter || pwm.Left.Green == 255){
            PWM_OUT_1 = HIGH;
        }
        else{
            PWM_OUT_1 = LOW;
        }

        if (pwm.Left.Blue > pwmCounter || pwm.Left.Blue == 255){
            PWM_OUT_2 = HIGH;
        }
        else{
            PWM_OUT_2 = LOW;
        }

        if (pwm.Right.Red > pwmCounter || pwm.Right.Red == 255){
            PWM_OUT_3 = HIGH;
        }
        else{
            PWM_OUT_3 = LOW;
        }

        if (pwm.Right.Green > pwmCounter || pwm.Right.Green == 255){
            PWM_OUT_4 = HIGH;
        }
        else{
            PWM_OUT_4 = LOW;
        }

        if (pwm.Right.Blue > pwmCounter || pwm.Right.Blue == 255){
            PWM_OUT_5 = HIGH;
        }
        else{
            PWM_OUT_5 = LOW;
        }

        if (pwm.Aux.Red > pwmCounter || pwm.Aux.Red == 255){
            PWM_OUT_6 = HIGH;
        }
        else{
            PWM_OUT_6 = LOW;
        }

        if (pwm.Aux.Green > pwmCounter || pwm.Aux.Green == 255){
            PWM_OUT_7 = HIGH;
        }
        else{
            PWM_OUT_7 = LOW;
        }

        if (pwm.Aux.Blue > pwmCounter || pwm.Aux.Blue == 255){
            PWM_OUT_8 = HIGH;
        }
        else{
            PWM_OUT_8 = LOW;
        }

        pwmCounter++;
    }
//******************************************************************************
// Gestione interrupt RECEIVE USART
//******************************************************************************
    if(RCIF)
    {
//Leggo il buffer seriale
        rx_chars[rx_counter] = RCREG;
//Incremento il contatore caratteri
        rx_counter++;
        if(rx_counter == 4){
            newCmd = 1;
        }
//NOTA: il flag di interrupt si azzera in automatico dopo la lettura
    }

}

//##############################################################################
// Main
//##############################################################################
void main(){
//******************************************************************************
// Inizializzazione
//******************************************************************************
    PORTS_SETUP();
    TIMER0_SETUP();
    ADC_SETUP();
    init_comms();
    INTERRUPT_SETUP();
//Impostazioni di default all'alimentazione del dispositivo
    power = OFF;
    operationMode = LOCAL_MODE;
    effectMode = AUTO;
    currentEffect = RAINBOW;
    effectMax = MAX_RAINBOW;
    TogglePower();
//******************************************************************************
// Inizio ciclo principale
//******************************************************************************
    while(TRUE){
        if (systemTick){
            systemTick = 0;
//Avvio la funzione di conversione e leggo pulsanti e potenziometro
            AdConversion();
//Verifico se vi  una nuova lettura dei pulsanti, in caso positivo controllo
//se un pulsante  premuto ed eseguo i comandi
            if(btnIsValid){
                btn = AnalogBtn(btnValue);
                switch (btn){
                    case EFFECT:
                        currentEffect++;
                        if(currentEffect == NUMBER_OF_EFFECTS){currentEffect=0;}
                        if(currentEffect == RAINBOW){effectMax=MAX_RAINBOW;}
                        if(currentEffect == SOFT){effectMax=MAX_SOFT;}
                        break;
                    case POWER:
                        power = !power;
                        TogglePower();
                        break;
                    case MODE:
                        effectMode = !effectMode;
                        break;
                }
            }

//Verifico se vi  una nuova lettura del potenziometro, in caso positivo eseguo
//i comandi
//Se la modalit  AUTO, il potenziometro modifica la durata dell'effetto
//Viceversa se MANUAL, il potenziometro modifica il colore
            if(potIsValid){
                potIsValid = 0;
                if (effectMode == AUTO){
                    autoTimer +=AUTOTIMER_INCREMENT;
                    if(autoTimer > (potValue)){
                        effectValue +=EFFECT_INCREMENT;
                        autoTimer = 0;
                        if (effectValue >= effectMax){effectValue = 0;}
                    }
                }
                else
                {
                    effectValue = 1023 - potValue;
                }
            }
//Imposto il duty cicle dei led e richiamo l'effetto se necessario
            switch (operationMode){
                case PWR_OFF:
                    pwm.Left.Red = 0;
                    pwm.Left.Green = 0;
                    pwm.Left.Blue = 0;
                    pwm.Right = pwm.Left;
                    pwm.Aux = pwm.Left;
                    break;
                case LOCAL_MODE:
                    if(currentEffect == SOFT){
                        effect = SoftRGB(effectValue);
                    }
                    if(currentEffect == RAINBOW){
                        effect = RainbowRGB(effectValue);
                    }
                    pwm.Left = effect;
                    pwm.Right = effect;
                    pwm.Aux = effect;
                    break;
                case PC_MODE:
                    if(currentEffect == SOFT){
                        effect = SoftRGB(effectValue);
                    }
                    if(currentEffect == RAINBOW){
                        effect = RainbowRGB(effectValue);
                    }
                    pwm.Aux = effect;
                    break;
            }

//Verifico se la comunicazione seriale si  interrotta (timeout)
            if(rx_counter > 0){com_timeout++;}
            if(com_timeout == TIMEOUT){
                rx_counter = 0;
                com_timeout = 0;
                SendTimeout;
            }

        }
//******************************************************************************
// Fine ciclo principale
//******************************************************************************
// Controllo comandi da seriale
//******************************************************************************
        if(newCmd){
            newCmd = 0;
            com_timeout = 0;
            PcCommands();
            if(operationMode == PC_MODE){
                pwm.Left = computer.Left;
                pwm.Right = computer.Right;
                pwm.Aux = effect;
            }
            if(operationMode == PC_MODE_WITH_AUX){
                pwm = computer;
            }
        }

    }
}

//##############################################################################
// Funzione di decodifica dei comandi da pc
//##############################################################################
void PcCommands(){
//Azzero il contatore caratteri, per poter ricevere nuovi comandi
    rx_counter = 0;
//Controllo il primo byte ricevuto, che contiene il comando
//Gli altri 3 byte contengono i parametri
    switch (OP_BYTE){
        case LEFT_BAR:
            computer.Left.Red = RED_BYTE;
            computer.Left.Green = GREEN_BYTE;
            computer.Left.Blue = BLUE_BYTE;
            break;
        case RIGHT_BAR:
            computer.Right.Red = RED_BYTE;
            computer.Right.Green = GREEN_BYTE;
            computer.Right.Blue = BLUE_BYTE;
            break;
        case AUX_BAR:
            computer.Aux.Red = RED_BYTE;
            computer.Aux.Green = GREEN_BYTE;
            computer.Aux.Blue = BLUE_BYTE;
            break;
        case COMMAND:
            if (CMD_BYTE <= 2 && power == ON){
                SendACK;
                operationMode = CMD_BYTE;
            }
            else if(CMD_BYTE == PWR_OFF){
                SendACK;
                power = OFF;
                TogglePower();
            }
            else if(CMD_BYTE == PWR_ON){
                SendACK;
                power = ON;
                TogglePower();
            }
            else{SendNACK;}
            break;
        case READ_AUTHOR:
            SendAuthor;
            break;
        case READ_VERSION:
            SendVersion;
            break;
        case READ_HWVERSION:
            SendHwVersion;
            break;
        case READ_SYSTEM:
            SendSystem;
            break;
        case READ_LAST_UPDATE:
            SendLastUpdate;
            break;
        default:
            SendNACK;
            break;
    }
}
//##############################################################################
// Funzione di acquisizione dati analogici
//##############################################################################
void AdConversion(){
    static unsigned char i;
    static unsigned char j;
    if (AdcOperation == 0){
        GO = 1;
        while(GO);
//memorizzo la lettura del potenziometro in un buffer circolare a 8 posizioni
        potBuffer[i] = ADC_RESULT;
        i++;
        if(i==8){i=0;}
//sommo tra di loro i valori nel buffer e divido per ottenere la lettura media
        potValue = 0;
        for(j=0; j<=7; j++){
            potValue += potBuffer[j];
        }
        potValue /= 8;
        potIsValid = TRUE;
        btnIsValid = FALSE;
        SELECT_ADC_CHANNEL(BTN_INPUT);
    }
    if (AdcOperation == 1){
        GO  = 1;
        while(GO);
        btnValue = ADC_RESULT;
        btnIsValid = TRUE;
        potIsValid = FALSE;
        SELECT_ADC_CHANNEL(POT_INPUT);
    }
//Incremento il contatore dello step di conversione
    AdcOperation++;
    if (AdcOperation == 2){AdcOperation = 0;}
}
//##############################################################################
// Funzione di gestione dei pulsanti analogici
//##############################################################################
unsigned char AnalogBtn(unsigned int btnValue){
    static unsigned char btnDebounceCounter;
    static unsigned char btnSelected;
    static bit cmdOk;

//Controllo se la lettura rientra nei valori previsti dai pulsanti, e se il
//comando non  gi stato eseguito
    if(btnValue >= BTN1_MIN && btnValue <= BTN1_MAX && !cmdOk){
//Se il pulsante  appena stato premuto lo imposto come selezionato e azzero
//il conteggio antirimbalzo
        if (btnSelected != BTN1){
            btnSelected = BTN1;
            btnDebounceCounter = 0;
        }
//Viceversa incremento il conteggio
        else{
            btnDebounceCounter ++;
        }
    }

    if(btnValue >= BTN2_MIN && btnValue <= BTN2_MAX && !cmdOk){
        if (btnSelected != BTN2){
            btnSelected = BTN2;
            btnDebounceCounter = 0;
        }
        else{
            btnDebounceCounter ++;
        }
    }

    if(btnValue >= BTN3_MIN && btnValue <= BTN3_MAX && !cmdOk){
        if (btnSelected != BTN3){
            btnSelected = BTN3;
            btnDebounceCounter = 0;
        }
        else{
            btnDebounceCounter ++;
        }
    }
//Trascorso il tempo di debounce, restituisco il pulsante premuto
    if (btnDebounceCounter == DEBOUNCE_COUNT){
//Azzero il contatore e setto il flag del comando eseguito
        btnDebounceCounter = 0;
        cmdOk = 1;
        return btnSelected;
    }
//Quando il pulsante viene rilasciato azzero il flag
    if (btnValue >= DEFAULT_VALUE){
        cmdOk = 0;
        return 0;
    }
}
//##############################################################################
// Gestione power ON/OFF
//##############################################################################
void TogglePower(){
//Spegnimento controller
    if (power == OFF){
//Salvataggio configurazione corrente
        GIE = FALSE;
        cfgVar = effectMode;
        eeprom_write(MODE_ADDRESS, cfgVar);
        cfgVar = currentEffect;
        eeprom_write(EFFECT_ADDRESS, cfgVar);
        GIE = TRUE;
//Spengo i led
        operationMode = PWR_OFF;
//Invio notifica su seriale
        SendPwrOff;
    }
//Accensione controller
    else{
//Carico cfg salvata
        GIE = FALSE;
        cfgVar = eeprom_read(MODE_ADDRESS);
        effectMode = cfgVar;
        cfgVar = eeprom_read(EFFECT_ADDRESS);
        currentEffect = cfgVar;
        GIE = TRUE;
//Abilito led in default (controllo locale)
        operationMode = LOCAL_MODE;
//Invio notifica su seriale
        SendPwrOn;
    }
}
//##############################################################################
// Effetti di luce
//##############################################################################
color SoftRGB(unsigned int analog){
    color softRgb;
    if(analog == 0){
        softRgb.Red = 0;
        softRgb.Green = 0;
        softRgb.Blue = 0;
    }
    if(analog >= 1 && analog <= 73){
        softRgb.Red = (analog * 35) /10;
        softRgb.Green = 0;
        softRgb.Blue = 0;
    }
    if(analog >= 74 && analog <= 146){
        softRgb.Red = 255 - (((analog - 73) * 35) /10);
        softRgb.Green = 0;
        softRgb.Blue = 0;
    }
    if(analog >= 147 && analog <= 219){
        softRgb.Red = ((analog - 146) * 35) /10;
        softRgb.Green = 0;
        softRgb.Blue = softRgb.Red;
    }
    if(analog >= 220 && analog <= 292){
        softRgb.Red = 255 - (((analog - 219) * 35) /10);
        softRgb.Green = 0;
        softRgb.Blue = softRgb.Red;
    }
    if(analog >= 293 && analog <= 365){
        softRgb.Red = 0;
        softRgb.Green = 0;
        softRgb.Blue = ((analog - 292) * 35) /10;
    }
    if(analog >= 366 && analog <= 438){
        softRgb.Red = 0;
        softRgb.Green = 0;
        softRgb.Blue = 255 - (((analog - 365) * 35) /10);
    }
    if(analog >= 439 && analog <= 511){
        softRgb.Red = 0;
        softRgb.Green = ((analog - 438) * 35) /10;
        softRgb.Blue = softRgb.Green;
    }
    if(analog >= 512 && analog <= 584){
        softRgb.Red = 0;
        softRgb.Green = 255 - (((analog - 511) * 35) /10);
        softRgb.Blue = softRgb.Green;
    }
    if(analog >= 585 && analog <= 657){
        softRgb.Red = 0;
        softRgb.Green = ((analog - 584) * 35) /10;
        softRgb.Blue = 0;
    }
    if(analog >= 658 && analog <= 730){
        softRgb.Red = 0;
        softRgb.Green = 255 - (((analog - 657) * 35) /10);
        softRgb.Blue = 0;
    }
    if(analog >= 731 && analog <= 803){
        softRgb.Red = ((analog - 730) * 35) /10;
        softRgb.Green = softRgb.Red;
        softRgb.Blue = 0;
    }
    if(analog >= 804 && analog <= 876){
        softRgb.Red = 255 - (((analog - 803) * 35) /10);
        softRgb.Green = softRgb.Red;
        softRgb.Blue = 0;
    }
    if(analog >= 877 && analog <= 949){
        softRgb.Red = ((analog - 876) * 35) /10;
        softRgb.Green = softRgb.Red;
        softRgb.Blue = softRgb.Red;
    }
    if(analog >= 950 && analog <= 1022){
        softRgb.Red = 255 - (((analog - 949) * 35) /10);
        softRgb.Green = softRgb.Red;
        softRgb.Blue = softRgb.Red;
    }
    return softRgb;
}

color RainbowRGB(unsigned int analog){
    color rainbowRgb;
    if (analog < 2)
    {
        rainbowRgb.Red = 255;
        rainbowRgb.Green = 0;
        rainbowRgb.Blue = 0;
    }
    if (analog >= 2 && analog <= 171)
    {
        rainbowRgb.Red = 255;
        rainbowRgb.Green = 0;
        rainbowRgb.Blue = ((analog - 1) * 15)/10;
    }
    if (analog >= 172 && analog <= 341)
    {
        rainbowRgb.Red = 255-(((analog - 171) * 15)/10);
        rainbowRgb.Green = 0;
        rainbowRgb.Blue = 255;
    }
    if (analog >= 342 && analog <= 511)
    {
        rainbowRgb.Red = 0;
        rainbowRgb.Green = ((analog - 341) * 15)/10;
        rainbowRgb.Blue = 255;
    }
    if (analog >= 512 && analog <= 681)
    {
        rainbowRgb.Red = 0;
        rainbowRgb.Green = 255;
        rainbowRgb.Blue = 255-(((analog - 511) * 15)/10);
    }
    if (analog >= 682 && analog <= 851)
    {
        rainbowRgb.Red = ((analog - 681) * 15)/10;
        rainbowRgb.Green = 255;
        rainbowRgb.Blue = 0;
    }
    if (analog >= 852 && analog <= 1021)
    {
        rainbowRgb.Red = 255;
        rainbowRgb.Green = 255;
        rainbowRgb.Blue = ((analog - 851) * 15)/10;
    }
    if (analog == 1022 || analog == 1023)
    {
        rainbowRgb.Red = 255;
        rainbowRgb.Green = 255;
        rainbowRgb.Blue = 255;
    }
    if(analog > 1023 && analog <= 1192)
    {
        rainbowRgb.Red = 255;
        rainbowRgb.Green = 255-(((analog - 1022) * 15)/10);
        rainbowRgb.Blue = rainbowRgb.Green;
    }
    return rainbowRgb;
}
/*******************************************************************************
 * End Of: RGB_Led_Controller.c
 * ********************************** ******************************************/